function [dLdW, dLdb] = backwardLoss(X,Y,Y_hat,W,b, ActivationFunction, LossCriterion)
% backwardLoss calculates the gradient of L with regard to W and b
    dLdphiList = struct(...
        'mae', @(Y,T) sign(Y-T),...
        'mse', @(Y,T) Y-T);
    dphidxList = struct(...
        'logistic', @(x) exp(x)./(1 + exp(x)).^2,...
        'tanh', @(x) 4 * exp(2*x)./(exp(2*x)+1).^2,...
        'relu', @(x) x>0);
    
    dLdphi = dLdphiList.(lower(LossCriterion));
    dphidx = dphidxList.(lower(ActivationFunction));
    dWTxdW = @(x) x;
    
    % zip W,b
    W = [W;b'];
    X = [X; ones(1,size(X,2))];
    
    n = size(X,2);
%     dLdW = zeros(size(W));
%     for i = 1:n % this for-loop can be accelarated
%         x     = X(:,i);
%         y     = Y(:,i);
%         y_hat = Y_hat(:,i);
%         
%         dLdW_i = dLdphi(y_hat, y) .* dphidx(W' * x) * dWTxdW(x);
%         dLdW = dLdW + dLdW_i;
%     end
    dLdW = dWTxdW(X) * (dLdphi(Y_hat, Y) .* dphidx(W' * X))';
    dLdW = dLdW/n;
    
    % unzip W,b
    dLdb = dLdW(end,:);
    dLdW = dLdW(1:end-1,:);
end